@@ -273,15 +273,13 @@ void *get_mmap_addr(unsigned long size)
 
 #else
 
-#ifdef _BSD
+#ifdef _WIN32
+#include <windows.h>
+#elif defined(_BSD)
 #include <stdlib.h>
 #else
 #include <malloc.h>
 #endif
-#ifdef _WIN32
-/* XXX: find a solution to have page aligned data */
-#define memalign(align, size) malloc(size)
-#endif
 
 int qemu_write(int fd, const void *buf, size_t n)
 {
@@ -308,8 +306,24 @@ void *qemu_malloc(size_t size)
     return malloc(size);
 }
 
-#if defined(USE_KQEMU)
+#if defined(_WIN32)
 
+void *qemu_vmalloc(size_t size)
+{
+    /* FIXME: this is not exactly optimal solution since VirtualAlloc
+       has 64Kb granularity, but at least it guarantees us that the
+       memory is page aligned. */
+    return VirtualAlloc(NULL, size, MEM_COMMIT, PAGE_READWRITE);
+}
+
+void qemu_vfree(void *ptr)
+{
+    VirtualFree(ptr, 0, MEM_RELEASE);
+}
+
+#elif defined(USE_KQEMU)
+
+#include <sys/vfs.h>
 #include <sys/mman.h>
 #include <fcntl.h>
 
@@ -320,11 +334,37 @@ void *qemu_vmalloc(size_t size)
     const char *tmpdir;
     char phys_ram_file[1024];
     void *ptr;
+    struct statfs stfs;
 
     if (phys_ram_fd < 0) {
         tmpdir = getenv("QEMU_TMPDIR");
         if (!tmpdir)
             tmpdir = "/dev/shm";
+        if (statfs(tmpdir, &stfs) == 0) {
+            int64_t free_space;
+            int ram_mb;
+
+            extern int ram_size;
+            free_space = (int64_t)stfs.f_bavail * stfs.f_bsize;
+            if ((ram_size + 8192 * 1024) >= free_space) {
+                ram_mb = (ram_size / (1024 * 1024));
+                fprintf(stderr, 
+                        "You do not have enough space in '%s' for the %d MB of QEMU virtual RAM.\n",
+                        tmpdir, ram_mb);
+                if (strcmp(tmpdir, "/dev/shm") == 0) {
+                    fprintf(stderr, "To have more space available provided you have enough RAM and swap, do as root:\n"
+                            "umount /dev/shm\n"
+                            "mount -t tmpfs -o size=%dm none /dev/shm\n",
+                            ram_mb + 16);
+                } else {
+                    fprintf(stderr, 
+                            "Use the '-m' option of QEMU to diminish the amount of virtual RAM or use the\n"
+                            "QEMU_TMPDIR environment variable to set another directory where the QEMU\n"
+                            "temporary RAM file will be opened.\n");
+                }
+                exit(1);
+            }
+        }
         snprintf(phys_ram_file, sizeof(phys_ram_file), "%s/qemuXXXXXX", 
                  tmpdir);
         if (mkstemp(phys_ram_file) < 0) {
